/*
** Module   : AbstractWidget.java
** Abstract : Base class for widgets.
**
** Copyright (c) 2010-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100629 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 SIY 20100822 Added base implementation of new interface methods config() and isFocusable().
** 004 SIY 20101122 Switching from KeyEvent to KeyInput.
** 005 SIY 20111115 Added physical widget location.
** 006 CA  20131121 The hidden/visible state changes must be collected when the widget's attribute
**                  is actually set.
** 007 EVL 20140617 Added method to check if the widget needs the cursor to be on.
** 008 CA  20140711 Enhanced font support. Added physicalDimension, in preparation for GUI support.
** 009 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
**                  Added mouse event processing.  Added APIs to work with physical units, 
**                  depending on the UI type.
** 010 MAG 20140912 Added insets member used on GUI only.
** 011 HC  20140911 GUI coordinate support - data types refactoring.
** 012 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 013 MAG 20140917 Added parent width and height inheritance.
**                  Added mouse position translate.
** 014 CA  20141021 Changed the resize() API to specify the X and Y deltas, plus flags to indicate
**                  if the top-left corner needs to move on the X or Y axis.
**                  Added API to allow a widget to raise (or not) legacy mouse events, when a 
**                  mouse action was performed on it; by default all widgets can raise these events.
** 015 VIG 20141016 Added window() method to resolving widgets Window.
** 016 MAG 20141020 Added clipping rectangle computation.
** 017 HC  20141021 Implemented window move/position support.
** 018 HC  20141118 Added screenPhysicalLocation().
** 019 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 020 HC  20141207 Implemented runtime support for widget coordinate attributes.
** 021 HC  20150111 Fixed widget position change in a frame.
** 022 HC  20150128 Added support for frame widget position changes induced by setting
**                  ROW/COL/X/Y attributes.
** 023 CA  20150212 Allow events to be captured and processed by the widgets without server-side
**                  or main event-loop notification.  This fixes problems in the EDITING and
**                  WAIT-FOR event loops, for i.e. COMBO-BOX and EDITOR widgets.
** 024 IAS 20150219 Added getVisiblePart() method
** 025 HC  20150323 Removed business logic from config classes.
** 026 CA  20150429 When creating a rectangle from a location and dimension, use the base units
**                  to determine the coordinates for current client mode.
** 027 VIG 20150330 Added prepParentLocation as extension point. Added pop-up menu key listener
**                  adding.
** 026 GES 20150507 Added a comment to clarify that AWT events usage is intentional.
** 027 HC  20150612 Implemented support for modal windows and GUI ALERT-BOX.
** 028 HC  20150626 Fixed an unhandled exception when showing a modal window.
** 029 HC  20150702 Improved parent-child relationship of top-level windows.
** 030 IAS 20150703 Added 'wasRealized' flag support, 'debug' method
*/

package com.goldencode.p2j.ui.client.widget;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.lang.ref.*;
import java.util.*;
import java.util.logging.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.Menu.PopupKeyListener;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.util.*;

/**
 * Base widget class. 
 */
public abstract class AbstractWidget<O extends OutputManager<?>>
implements Widget<O>
{
   /** Logger. */
   protected static final Logger LOG = LogHelper.getLogger(AbstractWidget.class.getName());

   /** Widget enabled flag. */
   private boolean enabled = true;
   
   /** Widget hidden flag. */
   private boolean hidden = false;
   
   /** Widget highlighted flag. */
   private boolean highlighted = false;
   
   /** Widget visible flag. This field is needed only for instances without config. */
   private boolean visible = true;
   
   /** Inherit parent width */
   private boolean inheritWidth = false;

   /** Inherit parent height */
   private boolean inheritHeight = false;

   /** Storage for focus listeners. */
   private ArrayList<FocusListener> focusListeners = new ArrayList<FocusListener>();
   
   /** Storage for key listeners. */
   private ArrayList<KeyListener> keyListeners = new ArrayList<KeyListener>();

   /** Reference to parent container. */
   private WeakReference<Container<O>> parent = null;
   
   /** <code>true</code> if and only if {@link PopupKeyListener} was added to this widget. */ 
   private boolean popupKeyListnrAdded = false;

   /** List of listeners which require notifications about the state of this widget. */
   private List<WidgetStateListener> listeners = new ArrayList<>(); 

   /** Reference to screen. */
   private O screen;
   
   /**
    * Widget origin (upper left corner) relative to container
    * expressed in character cell units.
    */
   protected Point location = new Point(0, 0);
   
   /**
    * Widget physical origin (upper left corner) relative to
    * container. expressed in units native to UI type.
    */
   protected NativePoint physicalLocation = new NativePoint(0, 0);
   
   /** Widget insets */
   protected Insets insets = Insets.ZERO_INSETS;
   
   /** Widget native insets */
   protected NativeInsets nativeInsets = NativeInsets.ZERO_INSETS;

   /**
    * Constructor
    */
   @SuppressWarnings("unchecked")
   public AbstractWidget()
   {
      screen = (O) OutputManager.instance();

      ClientConfigManager mgr = ConfigManager.getInstance();
      this.registerWidgetStateListener(mgr);
   }
   
   /**
    * Widget ID.
    * 
    * @return  see above.
    */
   @Override
   public WidgetId getId()
   {
      return UiUtils.getWidgetId(this);
   }

   /**
    * Get access to {@link OutputManager} instance.
    * 
    * @return  reference to {@link OutputManager}.
    */
   @Override
   public O screen() 
   {
      return screen;
   }

   /**
    * Add focus listener.
    * 
    * @param   listener
    *          Listener to add.
    */
   @Override
   public void addFocusListener(FocusListener listener)
   {
      focusListeners.add(listener);
   }

   /**
    * Add key listener.
    * 
    * @param   listener
    *          Listener to add.
    */
   @Override
   public void addKeyListener(KeyListener listener)
   {
      keyListeners.add(listener);
   }

   /**
    * Get ancestor window for this widget.
    * 
    * @return  ancestor window.
    */
   @SuppressWarnings("unchecked")
   @Override
   public TitledWindow<O> ancestor()
   {
      return parentOrSelf(TitledWindow.class);
   }
   
   /**
    * Get the ID of the ancestor window for this widget.
    * 
    * @return  ancestor window ID or <code>null</code> if no valid ancestor exists.
    */
   public WidgetId ancestorId()
   {
      Widget<?> ancestor = ancestor();
      if (ancestor == null)
      {
         return null;
      }
      
      return ancestor.getId();
   }

   /**
    * Get ancestor Window for this widget. This method must be called
    * only for widgets that are already attached to Window.
    * 
    * @return  ancestor Window.
    * 
    * @throws  IllegalStateException
    *          When this widget is not attached to window.
    */
   @Override
   public Window<O> window()
   {
      @SuppressWarnings("unchecked")
      Window<O> wnd = parentOrSelf(Window.class);
         
      if (!(wnd instanceof Window))
      {
         int wid = UiUtils.getWidgetIdAsInt(this);
         throw new IllegalStateException("Widget " + name() + " " + wid
                                         + " is not attached to a Window instance.");
      }
      
      return wnd;
   }
   
   /**
    * Get ancestor TopLevelWindow for this widget. This method must be called
    * only for widgets that are already attached to a TopLevelWindow.
    * 
    * @return ancestor TopLevelWindow.
    * 
    * @throws    IllegalStateException
    *            When this widget is not attached to a top-level window.
    */
   public TopLevelWindow<O> topLevelWindow()
   {
      @SuppressWarnings("unchecked")
      TopLevelWindow<O> wnd = parentOrSelf(TopLevelWindow.class);
         
      if (!(wnd instanceof TopLevelWindow))
      {
         int wid = UiUtils.getWidgetIdAsInt(this);
         throw new IllegalStateException("Widget " + name() + " " + wid
               + " is not attached to a TopLevelWindow instance.");
      }
      
      return wnd;
   }
   
   /**
    * Get widget bounding rectangle.
    * 
    * @return  bounding rectangle.
    */
   @Override
   public Rectangle bounds()
   {
      Rectangle result = new Rectangle(location(), dimension(), screen().coordinates().baseUnits());
      
      return result;
   }

   /**
    * Get widget insets.
    * 
    * @return  widget insets.
    */
   @Override
   public Insets insets()
   {
      return insets;
   }

   /**
    * Get widget dimension.
    * 
    * @return  widget dimension.
    */
   @Override
   public abstract Dimension dimension();

   /**
    * Get widget physical dimension, expressed in units native to UI type.
    *  
    * @return  widget dimension.
    */
   @Override
   public NativeDimension physicalDimension()
   {
      return screen.coordinates().dimensionToNative(dimension());
   }

   /**
    * Check if widget is displayed.
    * 
    * @return  <code>true</code> if widget is displayed.
    */
   @Override
   public boolean isDisplayed()
   {
      return (parent() == null) ? false : parent().isDisplayed();
   }

   /**
    * Draw widget.
    */
   @Override
   public abstract void draw();

   /**
    * Draw caret.
    */
   @Override
   public void drawCaret()
   {
   }

   /**
    * Check if widget is enabled.
    * 
    * @return  <code>true</code> if widget is enabled.
    */
   @Override
   public boolean isEnabled()
   {
      return enabled;
   }

   /**
    * Check if widget is focus traversable.
    * 
    * @return  <code>true</code> if widget is focus traversable.
    */
   @Override
   public boolean focusTraversable()
   {
      return (enabled && visible && !hidden);
   }

   /**
    * Force widget visibility change without associated repainting.
    * 
    * @param   visibility
    *          New visibility state.
    */
   @Override
   public void forceVisibility(boolean visibility)
   {
      _setVisible(visibility); // direct access
   }

   /**
    * Check if widget has focus.
    * 
    * @return  <code>true</code> if widget has focus.
    */
   @Override
   public boolean hasFocus()
   {
      return (parent() == null || parent().focus() != this) ? 
                                                         false : 
                                                         parent().hasFocus();
   }

   /**
    * Get widget height.
    * 
    * @return  widget height.
    */
   @Override
   public abstract double height();

   /**
    * Get current hidden state.
    * 
    * @return   hidden state.
    */
   @Override
   public boolean hidden()
   {
      return hidden;
   }

   /**
    * Hide widget.
    */
   @Override
   public void hide()
   {
      if (isVisible())
      {
         _setVisible(false); // direct access
         repaint();
      }
   }

   /**
    * Get current highlighted state.
    * 
    * @return   highlighted state.
    */
   @Override
   public boolean highlighted()
   {
      return highlighted;
   }

   /**
    * Get widget location relative to container.
    * <p>
    * The method always returns a valid location. If a component
    * has not been assigned location explicitly, the method returns
    * the default location 0,0.
    * 
    * @return  relative widget location.
    */
   @Override
   public Point location()
   {
      BaseConfig bc = UiUtils.getConfig(this, BaseConfig.class);
      
      if (bc == null)
      {
         double col = location.x;
         double row = location.y;
         
         if (col == BaseConfig.INV_COORD)
         {
            col = 0;
         }

         if (row == BaseConfig.INV_COORD)
         {
            row = 0;
         }
           
         return new Point(col, row);
      }
      
      return ConfigHelper.getEffectiveLocation(bc);
   }

   /**
    * Get widget physical location relative to container, expressed in units native to UI type.
    * <p>
    * The method always returns a valid location. If a component has not been assigned location
    * explicitly, the method returns the default location 0,0.
    * <p>
    * The value returned by the method is in native (also called physical) units. Native units 
    * should not be used by the layout logic but only by the output subsystems.
    * 
    * @return  relative widget location.
    */
   @Override
   public NativePoint physicalLocation()
   {
      BaseConfig bc = UiUtils.getConfig(this, BaseConfig.class);
      
      if (bc == null)
      {
         int x = physicalLocation.x;
         int y = physicalLocation.y;
         
         if (x == BaseConfig.INV_COORD)
         {
            x = 0;
         }

         if (y == BaseConfig.INV_COORD)
         {
            y = 0;
         }
           
         return new NativePoint(x, y);
      }
      
      if (screen().isChui())
      {
         // in chui the characters-native conversion is 1:1 
         Point p = ConfigHelper.getEffectiveLocation(bc);
         return new NativePoint((int) p.x, (int) p.y);
      }
      else
      {
         // in gui the pixels-native conversion is 1:1
         return ConfigHelper.getEffectivePixelLocation(bc);
      }
   }
   
   /**
    * Get widget minimum size.
    * 
    * @return  widget minimum size.
    */
   @Override
   public abstract Dimension minimumSize();

   /**
    * Get widget name.
    * 
    * @return  widget name.
    */
   @Override
   public String name()
   {
      WidgetConfig c = config();
      return c instanceof BaseConfig ? ((BaseConfig) c).name : "";
   }

   /**
    * Get widget parent.
    * 
    * @return  widget parent.
    */
   @Override
   public Container<O> parent()
   {
      return (parent == null) ? null : parent.get();
   }
   
   /**
    * Get widget parent of the specified type.
    * 
    * @param   clazz
    *          The parent type to return.
    * 
    * @return  widget parent.
    */
   @Override
   public <T extends Widget<O>> T parent(Class<T> clazz)
   {
      Widget<?> parent = parent();
      while(parent != null && !(clazz.isAssignableFrom(parent.getClass())))
      {
         parent = parent.parent();
      }
      
      @SuppressWarnings("unchecked")
      T result = (T) parent;
      
      return result;
   }

   /**
    * Get this or widget parent of the specified type.
    * 
    * @param   clazz
    *          The type to return.
    * 
    * @return  widget parent or this.
    */
   @Override
   @SuppressWarnings("unchecked")
   public <T extends Widget<O>> T parentOrSelf(Class<T> clazz)
   {
      if (clazz.isAssignableFrom(getClass()))
      {
         return (T) this;
      }
      
      return parent(clazz);
   }
   
   /**
    * Process focus event.
    * 
    * @param   event
    *          Event to process.
    */
   @Override
   public void processFocusEvent(FocusEvent event)
   {
      for (FocusListener listener : focusListeners)
      {
         event.dispatch(listener);
      }
   }

   /**
    * Process key event.
    * 
    * @param   event
    *          Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput event)
   {
      processKeyListeners(event);
   }

   /**
    * Process key listeners.
    * 
    * @param    event
    *           Keyboard event.
    */
   @Override
   public void processKeyListeners(KeyInput event)
   {
      for (KeyListener listener : keyListeners)
      {
         event.dispatch(listener);

         if (event.isConsumed())
            break;
      }
   }

   /**
    * Request widget repainting.
    */
   @Override
   public void repaint()
   {
      if (!isDisplayed())
         return;

      repaint(bounds());
   }

   /**
    * Request widget repainting.
    * 
    * Only the given update rectangle will be repainted (note that it is
    * clipped by all parent containers). No event is posted if the widget
    * itself or any of the parent containers are hidden.
    * 
    * @param    ur
    *           The update rectangle (relative to the origin of its parent
    *           Container) which will be repainted. Note that it is clipped by
    *           all parent containers.
    */
   @Override
   public void repaint(Rectangle ur)
   {
      if (!isDisplayed())
         return;

      Widget<O> cm = this;
      Container<O> ct = parent();

      // do not readjust coordinates if the root window is reached
      while (ct != null && !(ct instanceof TopLevelWindow))
      {
         if (!ct.isVisible())
         {
            return;
         }

         if (ct.clipping())
         {
            ur = ur.intersection(ct.getClipRect());

            if (ur == null || ur.empty())
            {
               return;
            }
         }

         cm = ct;
         ct = cm.parent();
         ur = ur.translate(prepParentLocation(cm));
      }

      EventManager.postEvent(new PaintEvent(this, ur));
   }

   /**
    * Request focus.
    */
   @Override
   public void requestFocus()
   {
      Widget<O> currentFocus = ancestor().currentFocus();
      
      if (currentFocus == this)
         return;
      
      if (currentFocus != null)
         EventManager.postEvent(new FocusEvent(currentFocus, EventType.FOCUS_LOST));

      EventManager.postEvent(new FocusEvent(this, EventType.FOCUS_GAINED));

      parent().setFocus(this);
      repaint();
   }

   /**
    * Request screen synchronization.  
    */
   @Override
   public void requestSync()
   {
      EventManager.postEvent(new SyncEvent(this));
   }

   /**
    * Get absolute location of widget in screen coordinates.
    * 
    * @return  absolute widget location.
    */
   @Override
   public Point screenLocation()
   {
      if (parent() == null) {
         throw new IllegalStateException("widget is not in the container");
      }

      return Coordinate.scale(parent().screenLocation().translate(location()));
   }

   /**
    * Get absolute location of widget in screen coordinates.
    * The result is in native units.
    * 
    * @return  absolute widget location.
    */
   @Override
   public NativePoint screenPhysicalLocation()
   {
      if (parent() == null) 
      {
         throw new IllegalStateException("widget is not in the container");
      }

      return parent().screenPhysicalLocation().translate(physicalLocation());
   }
   
   /**
    * Get display location of widget in screen coordinates.
    * The result is in native units.
    * 
    * @return  display widget location.
    */
   @Override
   public NativePoint displayPhysicalLocation()
   {
      TopLevelWindow<?> window = topLevelWindow();
      if (window == null)
      {
         return screenPhysicalLocation();
      }
      
      // window physical location represents a location on the display, 
      // whereas screen physical location is the offset from the parent window
      return window.physicalLocation().translate(screenPhysicalLocation());
   }

  /**
    * Enable/disable widget.
    * 
    * @param   enabled
    */
   @Override
   public void setEnabled(boolean enabled)
   {
      _setEnabled(enabled);

      repaint();
   }

   /**
    * Tracks the hidden state of the widget.
    * 
    * @param    hidden
    *           New hidden state.
    */
   @Override
   public void setHidden(boolean hidden)
   {
      _setHidden(hidden);
   }

   /**
    * Tracks the highlighted state of the widget.
    * 
    * @param    highlighted
    *           New highlighted state.
    */
   @Override
   public void setHighlighted(boolean highlighted)
   {
      this.highlighted = highlighted;
   }

   /**
    * Set widget relative location.
    * 
    * The x value corresponds to COLUMN while
    * y value corresponds to ROW 4GL attribute.
    * Note however that the values are expected to be
    * 0-based.
    * 
    * @param   x
    *          Widget X.
    * @param   y
    *          Widget Y.
    */
   @Override
   public void setLocation(double x, double y)
   {
      if (isVisible())
         repaint(); 
      
      BaseConfig bc = UiUtils.getConfig(this, BaseConfig.class);
      
      if (bc == null)
      {
         // although little hackish this will nicely take advantage
         // of the config automated conversion 
         bc = new BaseConfig();
      }
      
      bc.clientColumn = x;
      bc.clientRow = y;
      
      // When this has associated config, physicalLocation/location is not effectively used. 
      // Still, keep it synchronized for clarity - for example this may prevent 
      // confusion during debugging.
      physicalLocation.x = bc.clientX;
      physicalLocation.y = bc.clientY;
      
      location.x = bc.clientColumn;
      location.y = bc.clientRow;
      
      if (isVisible())
         repaint();
   }

   /**
    * Set widget relative location.
    * 
    * @param   origin
    *          Location to set.
    */
   @Override
   public void setLocation(Point origin)
   {
      setLocation(origin.x, origin.y);
   }

   /**
    * Set widget relative physical location.
    * 
    * @param   x
    *          Widget physical X.
    * @param   y
    *          Widget physical Y.
    */
   @Override
   public void setPhysicalLocation(int x, int y)
   {
      BaseConfig bc = UiUtils.getConfig(this, BaseConfig.class);
      
      if (bc == null)
      {
         // although little hackish this will nicely take advantage
         // of the config automated conversion 
         bc = new BaseConfig();
      }
      
      if (screen().isChui())
      {
         // in chui the characters-native conversion is 1:1
         bc.clientColumn = x;
         bc.clientRow = y;
      }
      else
      {
         // in gui the pixels-native conversion is 1:1
         bc.clientX = x;
         bc.clientY = y;
      }
      
      // When this has associated config, physicalLocation/location is not effectively used. 
      // Still, keep it synchronized for clarity - for example this may prevent 
      // confusion during debugging.
      physicalLocation.x = bc.clientX;
      physicalLocation.y = bc.clientY;
      
      location.x = bc.clientColumn;
      location.y = bc.clientRow;
   }
   
   /**
    * Determine the physical bounds, relative to screen, expressed in units native to UI type.
    * 
    * @return  See above.
    */
   @Override
   public NativeRectangle physicalBounds()
   {
      return new NativeRectangle(physicalLocation(), physicalDimension());
   }
   
   /**
    * Move this widget on the X and Y axis, by the specified deltas.  Values are specified in
    * units depending on the UI type.
    * 
    * @param   dx
    *          The X-axis delta.
    * @param   dy
    *          The Y-axis delta.
    */
   @Override
   public void moveBy(int dx, int dy)
   {
      // no-op by default
   }
   
   /**
    * Resize the widget to the specified dimension.  Values are specified in units depending on
    * the UI type.
    * 
    * @param    dx
    *           The X axis adjustment.
    * @param    dy
    *           The Y axis adjustment.
    * @param    moveX
    *           Flag indicating that the top-left corner needs to be moved on the X axis.
    * @param    moveY
    *           Flag indicating that the top-left corner needs to be moved on the Y axis.
    */
   @Override
   public void resize(int dx, int dy, boolean moveX, boolean moveY)
   {
      // no-op by default
   }
   
   /**
    * Set widget parent.
    * 
    * @param   container
    *          Parent to set.
    */
   @Override
   public void setParent(Container<O> container)
   {
      parent = (container == null) ? 
               null : 
               new WeakReference<Container<O>>(container);
   }

   /**
    * Hide/show widget.
    * 
    * @param   visible
    *          New visible state.
    */
   @Override
   public void setVisible(boolean visible)
   {
      if (visible)
         show();
      else
         hide();
   }

   /**
    * Show widget.
    */
   @Override
   public void show()
   {
      if (!isVisible())
      {
         _setVisible(true); // direct access
         repaint();
      }
   }

   /**
    * Check if widget is totally obscured.
    * 
    * @return  <code>true</code> if widget is totally obscured.
    */
   @Override
   public boolean isTotallyObscured()
   {
      return WindowManager.checkRectangle(bounds(), ancestor());
   }

   /**
    * Check if widget is visible.
    * 
    * @return  <code>true</code> if widget is visible.
    */
   @Override
   public boolean isVisible()
   {
      return visible;
   }
   
   /**
    * Check if widget is part of a header.
    * 
    * @return  <code>true</code> if widget is part of header.
    */
   @Override
   public boolean isHeader()
   {
      return config().header;
   }

   /**
    * Check if widget needs in cursor to be on when having the input focus.  By default this
    * method always returns <code>true</code> so it needs to be overwritten if other return value
    * is required for particular widget implementation class.
    * 
    * @return  <code>true</code> as the default value.
    */
   @Override
   public boolean forceCursorOnFocus()
   {
      return true;
   }

   /**
    * Get widget width.
    * 
    * @return  widget width.
    */
   @Override
   public abstract double width();

   /**
    * Disable widget.
    */
   protected void disable()
   {
      _setEnabled(false);
   }

   /**
    * Enable widget.
    */
   protected void enable()
   {
      _setEnabled(true);
   }
   
   /**
    * Instructs the widget that its size may have changed. The widget updates
    * its state from the associated config instance. It is the responsibility
    * of the caller to call this method only after the config size is set with
    * the new values.
    * <p>
    * This implementation uses the supplied config instance to detect whether
    * size has changed. If <code>null</code> is passed in the method does
    * nothing.
    * 
    * @param    beforeUpdate
    *           {@link BaseConfig} reference or <code>null</code>.
    */
   protected void updateSizeWorker(BaseConfig beforeUpdate)
   {
      if (beforeUpdate == null)
      {
         return;
      }
      
      ControlConfig bc = UiUtils.getConfig(this, ControlConfig.class);
      
      // apply only to controls
      if (bc != null && beforeUpdate.widthChars != bc.widthChars)
      {
         repaint();
      }
   }   

   /**
    * Process event and dispatch it to specific method.
    * 
    * @param   evt
    *          Event to process.
    */
   @Override
   public void processEvent(Event evt)
   {
      if (evt instanceof KeyInput)
      {
         Widget<O> ancestor = ancestor();

         if (ancestor != null)
            ancestor.processKeyEvent((KeyInput) evt);
      }
      else if (evt instanceof FocusEvent)
         processFocusEvent((FocusEvent) evt);
   }

   /**
    * Get externalizable widget config which can be used to restore widget 
    * state after via network. 
    * 
    * Note that only very limited set of widgets actually have config.
    * By default widget has no config.
    * 
    * @param   <W>
    *          Config type.
    *          
    * @return  Reference to widget config.
    */
   @Override
   public <W extends WidgetConfig> W config()
   {
      return null;
   }
   
   /**
    * Checks whether this widget can receive focus generally.
    * 
    * @return  <code>true</code> if this widget can be focused under
    *          certain conditions. <code>false</code> if, and only if, this
    *          widget can never receive focus.    
    */
   @Override
   public boolean isFocusable()
   {
      return true;
   }
   
   /**
    * This API is responsible for resolving the font associated with this widget's title.  
    * This can be one of the default fonts, a special system font, or an entry in the font-table.
    * 
    * @return   See above.
    */
   public int resolveTitleFont()
   {
      throw new RuntimeException("Not implemented.");
   }
   
   /**
    * This API is responsible for resolving the widget's font.  This can be one of the default 
    * fonts, a special system font, or an entry in the font-table.
    * 
    * @return   See above.
    */
   public int resolveFont()
   {
      throw new RuntimeException("Not implemented.");
   }
   
   /**
    * Find the widget positioned just bellow the specified mouse position (in physical units).
    * 
    * @param    p
    *           The mouse physical location.
    * 
    * @return   A widget over which the mouse is positioned or <code>null</code> if none found.
    */
   @Override
   public Widget<O> findMouseSource(NativePoint p)
   {
      return isVisible() && this.physicalBounds().contains(p) ? this : null;
   }
   
   /**
    * Notification of a mouse dragged event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseDragged(MouseEvent e)
   {
      // no-op
   }

   /**
    * Notification of a mouse moved event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseMoved(MouseEvent e)
   {
      // no-op
   }

   /**
    * Notification of a mouse clicked event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      // no-op
   }

   /**
    * Notification of a mouse pressed event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      // no-op
   }

   /**
    * Notification of a mouse released event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      // no-op
   }

   /**
    * Notification of a mouse entered event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseEntered(MouseEvent e)
   {
      // no-op
   }

   /**
    * Notification of a mouse exited event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseExited(MouseEvent e)
   {
      // no-op
   }

   /**
    * Notification of a mouse wheel moved event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseWheelMoved(MouseWheelEvent e)
   {
      // no-op
   }
   
   /**
    * Destroy this widget.  Notify all listeners about the widget being destroyed.
    */
   public void destroy()
   {
      for (WidgetStateListener l : listeners)
      {
         l.notifyWidgetDestroyed(this);
      }
   }

   /**
    * Return whether width is inherited from parent widget.
    * 
    * @return  see above.
    */
   public boolean isInheritWidth()
   {
      return inheritWidth;
   }

   /**
    * Set if width is inherited from parent widget.
    * 
    * @param   inheritWidth
    *          <code>true</code> if width is inherited from parent. 
    */
   public void setInheritWidth(boolean inheritWidth)
   {
      this.inheritWidth = inheritWidth;
   }

   /**
    * Return whether height is inherited from parent widget.
    * 
    * @return  see above.
    */
   public boolean isInheritHeight()
   {
      return inheritHeight;
   }

   /**
    * Set if width is inherited from parent widget.
    * 
    * @param   inheritHeight
    *          <code>true</code> if height is inherited from parent. 
    */
   public void setInheritHeight(boolean inheritHeight)
   {
      this.inheritHeight = inheritHeight;
   }

   /**
    * Register a widget state listener for this widget.
    * 
    * @param    l
    *           The widget state listener.
    */
   public void registerWidgetStateListener(WidgetStateListener l)
   {
      if (l != null)
      {
         this.listeners.add(l);
      }
   }
   
   /**
    * Translate mouse position relative to widget. Widget location is relative to container on
    * which is layout. The mouse position is relative to window content pane which is the mouse
    * listener. This function translate mouse position coordinates relative to widget.
    *  
    * @param   widget
    *          Widget instance.
    * @param   evt
    *          Mouse event.
    *          
    * @return  see above.
    */
   public NativePoint translate(Widget<O> widget, MouseEvent evt)
   {
      // get widget physical location wich is relative to it container. 
      NativePoint p = widget.physicalLocation();

      // get inner containers and calculate position
      Widget<O> parent = widget.parent();
      
      while (!(parent instanceof TopLevelWindow<?>))
      {
         p.translate(parent.physicalLocation());
         parent = parent.parent();
      }
      
      // mouse position relative to widget position
      return new NativePoint(evt.getX() - p.x, evt.getY() - p.y);
   }   

   /**
    * Gets the current state of this component's data.
    * 
    * @return   The state of this widget's data  (for example 
    *           <code>ScreenBuffer.UNINITIALIZED</code>).
    */
   @Override
   public byte getState()
   {
      return config().state;
   }

   /**
    * Gets the sub-rectangle which should be visible.
    * 
    * @return   The sub-rectangle which should be visible 
    */
   public Rectangle getVisiblePart()
   {
      return null;
   }

   /**
    * Stores the new state of the component's data. 
    * 
    * @param    state
    *           The new state (for example <code>ScreenBuffer.CHANGED</code>).
    */
   @Override
   public void setState(byte state)
   {
      if (state < ScreenBuffer.STATE_MIN || state > ScreenBuffer.STATE_MAX)
         return;

      WidgetConfig config = this.config();
      if (config != null)
      {
         config.state = (byte) state;
      }
   }

   /**
    * Build widget clipped rectangle.
    * 
    * @return  see above.
    */
   protected NativeRectangle clipRectangle()
   {
      NativePoint origin = physicalLocation();
      NativeDimension dimension = physicalDimension();
      return clipRectangle(origin, dimension);
   }
   
   /**
    * Build clipped rectangle for given coordinates.
    * 
    * @param   origin
    *          Physical location.
    * @param   dimension
    *          Physical dimension.
    * 
    * @return  see above.
    */
   protected NativeRectangle clipRectangle(NativePoint origin, NativeDimension dimension)
   {
      // corrections for scrolling 
      int x = dimension.width  - (origin.x < 0 ? origin.x : 0);
      int y = dimension.height - (origin.y < 0 ? origin.y : 0);
      // clip rectangle
      return new NativeRectangle(0, 0, y, x);     
   }
   
   /**
    * Update the {@link #enabled} flag in the configuration and in the widget's state.
    * 
    * @param    enabled
    *           The widget's enabled state.
    */
   private void _setEnabled(boolean enabled)
   {
      this.enabled = enabled;
      if (config() != null)
      {
         config().enabled = enabled;
      }
   }
   
   /**
    * Update the {@link #visible} flag in the configuration and in the widget's state.
    * 
    * @param    visible
    *           The widget's visible state.
    */
   public void _setVisible(boolean visible)
   {
      this.visible = visible;
      WidgetConfig config = config();
      if (config != null)
      {
         config.visible = visible;
         if (visible && !config.realized)
         {
            config.realized = true;
            config.wasRealized = true;
         }
      }
   }
   
   /**
    * Update the {@link #hidden} flag in the configuration and in the widget's state.
    * 
    * @param    hidden
    *           The widget's hidden state.
    */
   private void _setHidden(boolean hidden)
   {
      this.hidden = hidden;
      if (config() != null)
      {
         config().hidden = hidden;
      }
   }

   /**
    * Check if the widget implementation can raise legacy, 4GL-style, mouse events.
    * 
    * @return   Always <code>true</code>.
    */
   @Override
   public boolean raiseLegacyMouseEvents()
   {
      return true;
   }
   
   /**
    * Determine the widget's window and synchronize output with physical terminal.
    * 
    * @see   OutputManager#sync(int)
    */
   public void sync()
   {
      int wid = topLevelWindow().getId().asInt();
      
      screen().sync(wid);
   }
   
   /**
    * Determine the widget's window and mark a rectangle as needing repainting. 
    * 
    * @param   rect
    *          rectangle to mark for repainting
    *          
    * @see     OutputManager#refresh(int, Rectangle)
    */
   public void refresh(Rectangle rect)
   {
      int wid = topLevelWindow().getId().asInt();
      
      screen().refresh(wid, rect);
   }
   
   /**
    * Determine the widget's window and clear specified area of the screen.
    * 
    * @param   origin
    *          Upper left corner of the area to clear.
    * @param   size
    *          Dimensions of the area.
    * @param   color
    *          Color.
    * 
    * @see     OutputManager#clear(int, NativePoint, NativeDimension, Color)
    */
   public void clear(Point origin, Dimension size, Color color)
   {
      int wid = topLevelWindow().getId().asInt();
      
      screen().clear(wid, origin, size, color);
   }
   
   /**
    * Determine the widget's window, get a copy of the current screen bitmap and reset clipping 
    * entirely. All screen positions become available for output. The previous state can be
    * restored using setBitmap() method.
    * 
    * @return  a copy of the current screen bitmap.
    * 
    * @see     OutputManager#getAndResetBitmap(int)
    */
   public ScreenBitmap getAndResetBitmap()
   {
      int wid = topLevelWindow().getId().asInt();

      return screen().getAndResetBitmap(wid);
   }
   
   /**
    * Determine the widget's window and reset the screen bitmap.
    * 
    * @see     OutputManager#resetBitmap(int)
    */
   public void resetBitmap()
   {
      int wid = topLevelWindow().getId().asInt();

      screen().resetBitmap(wid);
   }

   /**
    * Determine the widget's window, get a copy of the current screen bitmap and clip output 
    * further so it is guaranteed not to go outside of the provided rectangle.
    * 
    * @param   rect
    *          The extra clipping rectangle. <code>null</code> has no effect
    *          on clipping.
    * 
    * @return  a copy of the current screen bitmap.
    * 
    * @see     OutputManager#getAndSetBitmap(int, Rectangle)
    */
   public ScreenBitmap getAndSetBitmap(Rectangle rect)
   {
      int wid = topLevelWindow().getId().asInt();

      return screen().getAndSetBitmap(wid, rect);
   }

   /**
    * Determine the widget's window and get a copy of the current screen bitmap.
    * 
    * @return  a copy of the current screen bitmap.
    * 
    * @see     OutputManager#getBitmapCopy(int)
    */
   public ScreenBitmap getBitmapCopy()
   {
      int wid = topLevelWindow().getId().asInt();

      return screen().getBitmapCopy(wid);
   }
   
   /**
    * Determine the widget's window and restore the bitmap to a previous state.
    * 
    * @param   bitmap
    *          The bitmap instance to set as the output primitive's bitmap.
    * 
    * @see     OutputManager#setBitmap(int, ScreenBitmap)
    */
   public void setBitmap(ScreenBitmap bitmap)
   {
      int wid = topLevelWindow().getId().asInt();

      screen().setBitmap(wid, bitmap);
   }

   /**
    * Determine the widget's window and add a new clipping rectangle to the bitmap. If the 
    * rectangle is null, fill the entire bitmap with <code>true</code> values. Else, if the bitmap
    * was just reset, fill it with <code>false</code> values and set the clipping region to true.
    * 
    * @param   rect
    *          The clipping rectangle to be added.
    * 
    * @see     OutputManager#clipWith(int, Rectangle)
    */
   public void clipWith(Rectangle rect)
   {
      int wid = topLevelWindow().getId().asInt();

      screen().clipWith(wid, rect);
   }
   
   /**
    * Determine the widget's window and get the number of rows on the screen.
    * 
    * @return  The number of rows (lines of characters).
    */
   public double screenHeight()
   {
      int wid = topLevelWindow().getId().asInt();

      return screen.screenHeight(wid);
   }

   /**
    * Determine the widget's window and get the number of columns on the screen.
    *
    * @return  The number of columns (in characters).
    */
   public double screenWidth()
   {
      int wid = topLevelWindow().getId().asInt();

      return screen.screenWidth(wid);
   }

   /**
    * Determine the widget's window and get the screen dimensions.
    * 
    * @return screen dimension.
    */
   public Dimension screenDimension()
   {
      int wid = topLevelWindow().getId().asInt();

      return screen.screenDimension(wid);
   }
   
   /**
    * Check if the current widget can manage non-global system actions on its own, and not via the
    * enclosing event loop.
    * 
    * @return   <code>true</code> if the widget has its own event management.
    *           <code>false</code> by default.
    */
   public boolean manageSystemActions()
   {
      return false;
   }

   /**
    * Get the default event associated for this widget and the given key.
    * 
    * @param    key
    *           The current key.
    * @param    inEditingBlock
    *           <code>true</code> if EDITING mode is active.
    *           
    * @return   The action code for the widget's default event.
    */
   public int getDefaultEvent(KeyInput key, boolean inEditingBlock)
   {
      return -1;
   }

   /**
    * Process the action associated with the given key in an independent drawing bracket.
    * <p>
    * This must be called only in the context of widget-level event processing.
    * 
    * @param    key
    *           The key to be processed.
    */
   public void processSystemKey(final KeyInput key)
   {
      if (!manageSystemActions())
      {
         return;
      }
      
      // let the widget process this key
      ThinClient.getInstance().independentEventDrawingBracket(this, new Runnable()
      {
         @Override
         public void run()
         {
            processKeyEvent(key);
         }
      });
   }
   
   /**
    * Check if the given event needs to be ignored, as it can't be treated by the widget.
    * 
    * @param    action
    *           The event code.
    *           
    * @return   See above.
    */
   public boolean ignoreAction(int action)
   {
      return EventDefinition.isEventUniversal(action) ||
             action == Keyboard.KA_HELP               ||
             action == Keyboard.KA_TAB                ||
             action == Keyboard.SE_NEXT_FRAME         ||
             action == Keyboard.SE_PREV_FRAME
             ;
   }

   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    *           Config reference capturing the config state before 
    *           the owner's config was modified.
    */
   public void afterConfigUpdateBase(WidgetConfig beforeUpdate)
   {
      WidgetConfig c = config();
      
      if (c == null)
      {
         return;
      }
      
      setVisible(c.visible);
      setHidden(c.hidden);
      if (c.enabled != isEnabled())
      {
         setEnabled(c.enabled);
      }
      
      if (beforeUpdate instanceof BaseConfig)
      {
         updateSizeWorker((BaseConfig) beforeUpdate);
      }

      //  c.popupMenuId != beforeUpdate.popupMenuId don't work here 
      if (c.popupMenuId != -1 && !popupKeyListnrAdded)
      {
         addKeyListener(new Menu.PopupKeyListener());
         popupKeyListnrAdded = true;
      }
   }
   
   /**
    * Utility method for debugging, can be used to include information from the server-side
    * to the client-side log
    * 
    * @param message
    *       String parameter
    */
   public void debug(String message)
   {
      LOG.log(Level.INFO,  "{0};  #{1}: {2}", 
            new Object[] {this.getClass().getSimpleName(), getId(), message});
   }

   /**
    * Trick, that lets to correct parent location in cases, when it needed,
    * for example for menus.
    * 
    * @param    parent
    *           Parent component.
    *           
    * @return   corrected parent location.
    */
   protected Point prepParentLocation(Widget<?> parent)
   {
      return parent.location();
   }
}
